;.vcl=>.vsm : "masp -P . vu1.vcl | vcl_14beta7 -ovu1.vsm"
;.vsm=>.o   : "/usr/local/ps2dev/dvp/bin/dvp-as vu1.vsm  -o vu1.o"
;In case of strange behaviour verify that macros and [E]
;haven't vanished in .vsm (avoid comments right after macros,
;try to change order of instructions to get out of trouble).

.global vu1_start
.global vu1_end

.include "vcl_sml.i"

.init_vf_all
.init_vi_all
.syntax new

.vu

;per object stuff (per object stuff replaces per frame stuff)
LocalScreen	.assign 0	;rotation-translation matrix
LocalLight	.assign 4	;rotation matrix
LightDirection	.assign 8	;light direction
LightColor	.assign 9	;light color
Scales		.assign 10	;scales vector
Decals		.assign 11	;decals vector

;per frame stuff (valid at beginning of frame only)
FrameGifPacket	.assign 16	;GifPacket sent at beginning of frame

;per batch stuff
DataOffset	.assign 0	;Base for following data:

;input buffers (85 vertices, i.e triple qwords uv+norm+xyz)
NumberIn	.assign 0	;input : number of vertices
GifPacketIn	.assign 0	;input : Gif-Tag (integer number of vertices +0x8000 at .x)
UVIn		.assign 1	;input : floating point 2D texture coordinates
NormalIn	.assign 2	;input : floating point 3D normal vector
CoordsIn	.assign 3	;input : floating point 3D coordinates

;output buffers (85 vertices, i.e triple qwords uv+norm+xyz)
GifPacketOut	.assign 256	;output: Gif-Tag
UVOut		.assign 257	;output: integer<<4 2D coordinates
ColorOut	.assign 258	;output: integer color
CoordsOut	.assign 259	;output: integer<<4 2D coordinates
	
--enter
--endenter

vu1_start:

	;per frame stuff
	;clear frame and ZBuffer
	iaddiu	iKick, vi00, FrameGifPacket	;iKick=0+FrameGifPacket
	xgkick	iKick				;send Gif-Tag to GS (PS2 GPU)

	;per object stuff
	MatrixLoad mLocalScreen, LocalScreen, vi00
	MatrixLoad mLocalLight, LocalLight, vi00
	VectorLoad vLightDirection, LightDirection, vi00
	VectorLoad vLightColor, LightColor, vi00
	VectorLoad vScales, Scales, vi00
	VectorLoad vDecals, Decals, vi00

	fcset	0				;for clipping (resets previous n-1 n-2 etc... states)

--cont

begin:
	;draw batch

	;only one of these two lines must be uncommented at a time
	xtop	iBaseOffset			;quad buffering technic
	;iaddiu	iBaseOffset, vi00, DataOffset	;double buffering technic

	;read number of vertices
	ilw.x	iNumber, NumberIn(iBaseOffset)
	iadd	iTail, iNumber, iNumber		;x2
	iadd	iTail, iTail, iNumber		;x3
	iadd	iTail, iTail, iBaseOffset	;iTail=iBaseOffset+3*iNumber
	
	iadd	iPtr, vi00, iBaseOffset

	lq	iPacket, GifPacketIn(iBaseOffset)
	sq	iPacket, GifPacketOut(iBaseOffset)
	iaddiu	iNumber, iNumber, 0x7fff
	iaddiu  iNumber, iNumber, 1
	isw.x	iNumber, GifPacketOut(iBaseOffset)

loop:
--LoopCS 6,6	; tell VCL that we loop at least 6 times (provokes an incredible optimization!)

	VertexLoad vCoords, CoordsIn, iPtr
	;rotations, clipping scales, translation:
	MatrixMultiplyVertexW1 vCoords, mLocalScreen, vCoords
	;clipping: 
	clipw.xyz vCoords, vCoords
	fcand	vi01, 0x3FFFF			;vi01=1 if any vertice n, n-1 or n-2 is out of the area in front of camera
	iaddiu	iADC, vi01, 0x7FFF		;iADC=0x8000 if vi01=1 (p154 EE user manual)
	;forced 'no draw' (force consecutive independant triangles in case of triangle strips)
	ilw.w	iNoDraw, UVIn(iPtr)
	iadd	iADC, iADC, iNoDraw		;iADC>=0x8000 (skip drawing) if NoDraw flag set or if any vertice n, n-1 or n-2 is out
	;3D projection:
	VertexPersCorr vCoords, vCoords
	mul vCoords, vCoords, vScales	;final 2D scales
	add vCoords, vCoords, vDecals	;final 2D offsets
	VertexFpToGsXYZF2 vCoords, vCoords
	sq.xyz vCoords, CoordsOut(iPtr) ;can't use "VertexSave vCoords, CoordsOut, iPtr" here. Insruction order not respected by optimization.
	isw.w	iADC, CoordsOut(iPtr)

	VertexLoad vNormal, NormalIn, iPtr
	;rotation
	MatrixMultiplyVertexW1 vNormal, mLocalLight, vNormal
	VectorDotProduct fIntensity, vNormal, vLightDirection
	max.x  fIntensity, fIntensity, vf00[x]
	mul vColor, vLightColor, fIntensity[x]
	ColorFPtoGsRGBAQ vColor, vColor
	VertexSave vColor, ColorOut, iPtr

	VertexLoad vUV, UVIn, iPtr
	VertexSave vUV, UVOut, iPtr

	iaddiu	iPtr, iPtr, 3
	ibne	iPtr, iTail, loop		;loop until all of the verts in this batch are done.

	iaddiu	iKick,iBaseOffset,GifPacketOut	;iKick=0+iGifPacketOut+iBaseOffset
	xgkick	iKick				;send primitive to GS (PS2 GPU)
	
--cont
	b begin
--exit
--endexit

vu1_end:

.end
